package provide ::Aldec::NCsimConverter 1.0

if { [string equal $::tcl_platform(platform) windows] } {
	set __ncsimScriptsDirectory__ [file join [file dirname [info nameofexecutable]] .. scripts aldec ncsim_translator]
} else {
	set __ncsimScriptsDirectory__ [file join [file dirname [info nameofexecutable]] .. .. scripts aldec ncsim_translator]
}
 
source [file join $__ncsimScriptsDirectory__ token.tcl]
source [file join $__ncsimScriptsDirectory__ analyzers.tcl]
source [file join $__ncsimScriptsDirectory__ tokens.tcl]
source [file join $__ncsimScriptsDirectory__ replacers.tcl]
source [file join $__ncsimScriptsDirectory__ argument.tcl]
source [file join $__ncsimScriptsDirectory__ arguments.tcl]
unset __ncsimScriptsDirectory__

namespace eval ::Aldec::NCsimConverter {

	namespace export \
		convert

	#converts file _from into file _to
	proc convert { _from _to } {
	
		set fFrom [open $_from "r"]
		set fTo [open $_to "w"]
		set globalIndentation 0

		set tmpIndent ""

		puts "Processing..." 
		
		#header include
		puts $fTo "alib work"
		while { ![eof $fFrom] } {
			foreach line [getBunchOfLinesFromFile fFrom] {
				
				removeSurroundingEmptySpaces line				
				
				#ignore commentary, and convert '-' to '#'
				if { [string index $line 0] == "#" || [string index $line 0] == "-" } {
				
					if { [string index $line 0] == "-" } {
						set line [string replace $line 0 0 "#"]
					}
					
					puts $fTo "${tmpIndent}$line"
					continue
				}
				
				#convert line
				set indentation 0
				
				analyzeLine line indentation
				
				#prepare indentation --------------------
				#negative indentation apply for current line
				#positive indentation apply for next line
				
				set tmpIndent ""
				
				if { $indentation < 0 } {
					incr globalIndentation $indentation
				}
				
				for { set i 0 } { $i < $globalIndentation } { incr i } {
					#you can change indentation chars here, for example into few spaces instead of tab
					append tmpIndent "\t"
				}
								
				if { $indentation > 0 } {
					incr globalIndentation $indentation
				}
				#-----------------------------------------
				
				foreach line [splitFileLineBecauseOfSemicolon $line] {
					puts $fTo "${tmpIndent}$line"
				}
			}
		}	
		
		close $fFrom
		close $fTo

		puts "Script converted succesfully."

	}
	
	
#------------------------------ private -------------------------------------------	
	
	#returns complete line: it analyzes multilines, semicolons etc.
	#it can return few lines!!
	proc getBunchOfLinesFromFile { _fileId } {
		upvar $_fileId fileId
		
		set escapedBackslash 0
		
		#read multilines
		if { ![eof $fileId] } {
			gets $fileId line
			
			#check not-escaped backslash at the line end
			while { ![eof $fileId] && [string index $line end] == "\\" } {
				#check if backslash is escaped (extended identifier)
				if { [string index $line [expr [string length $line] -2]] == "\\" } {
					set escapedBackslash 1
					break
				}
				
				#replace backslash with single space and remove all white spaces
				#from next line (see ADZ1505)
				set line [string replace $line end end " "]
				set tmpLine [gets $fileId]
				removeSurroundingEmptySpaces tmpLine
				#
				
				append line $tmpLine
			}
		}
		
		if { !$escapedBackslash && [string index $line end] == "\\" } {
			set line [string replace $line end end " "]
		}
		
		return [splitFileLineBecauseOfSemicolon $line]
	}
	
	proc splitFileLineBecauseOfSemicolon { _line } {
		set lines {}
		set quotation 0
		set brace 0
		set idx 0
		set maxIdx [expr [string length $_line] - 1]
		set tmpLine ""
		while { $idx <= $maxIdx } {
			set char [string index $_line $idx]
			
			#check if semicolon is inside quotation
			if { $char == "\"" } {
				if { $idx == 0 } {
					set quotation [expr !$quotation]
				} else {
					#ignore escaped quotation
					set charBefore [string index $_line [expr $idx - 1]]
					if { $charBefore != "\\" } {
						set quotation [expr !$quotation]
					}
				}
			}
			
			if { $char == ";" && !$quotation } {								
				removeSurroundingEmptySpaces tmpLine
				lappend lines $tmpLine
				set tmpLine ""				
			} else {
				append tmpLine $char
			}
			incr idx
		}
		
		removeSurroundingEmptySpaces tmpLine
		lappend lines $tmpLine

		return $lines
	}
	
	proc removeSurroundingEmptySpaces { _line } {
		upvar $_line line
		regexp {^[[:blank:]]*(.*)[[:blank:]]*$} $line _tmp_ line
	}
	
	proc analyzeLine { _line _indentation } {
	
		upvar $_line line
		upvar $_indentation indentation
		
		set tokens [new Tokens]
	
		Tokens::clear $tokens
	
		set idx 0
		set maxIdx [expr [string length $line] - 1]

		while { $idx <= $maxIdx } {
			set token [getToken idx maxIdx line]
			Tokens::add $tokens $token
		}
		
		#change tokens according to our rules
		if { [catch {	analyzeTokens $tokens } errorMsg] } {
			error "Cannot analyze line '$line'. $errorMsg"
		}
				
		#join tokens to string
		getJoinedTokens $tokens line indentation
		
		delete $tokens
	}
	
	proc analyzeTokens { _tokens } {
	
		#[DEBUG]
		#foreach token [Tokens::getTokensList $_tokens] {			
			#puts "token = [Token::getString $token]"
			#puts "	isEmpty = [Token::isEmpty $token]"
			#puts "	isString = [Token::isString $token]"
			#puts "	isOpenBrace = [Token::isOpenBrace $token]"
			#puts "	isCloseBrace = [Token::isCloseBrace $token]"
			#puts "	isConverted = [Token::isConverted $token]"
		#}	
	
		while { [Tokens::next $_tokens token] } {
			set tokenString [Token::getString $token]
			
			if [isKnownCommand $tokenString] {
				eval Analyzers::$tokenString $_tokens
			}
		}
	}
	
	proc getJoinedTokens { _tokens _line _indentationCounter } {
		upvar $_line line
		upvar $_indentationCounter indentationCounter
	
		set line ""
		
		foreach token [Tokens::getTokensList $_tokens] {
			
			if { [Token::isEmpty $token] } {

				#append line " "
				append line [Token::getString $token]

			} elseif { [Token::isString $token] } {

				append line [Token::getString $token]

			} elseif { [Token::isOpenBrace $token] } {

				incr indentationCounter
				append line [Token::getString $token] 			

			} elseif { [Token::isCloseBrace $token] } {

				incr indentationCounter -1
				append line [Token::getString $token]
				
			}
		}		
	}	
	
	proc getToken { _idx _maxIdx _line } {
		
		upvar $_idx idx
		upvar $_maxIdx maxIdx
		upvar $_line line
		
		set char [string index $line $idx]
		
		#we have three kind of tokens:
		#1) empty space token
		#2) open/close brace
		#3) everything else as string
		#	- get chars until non-escaped space
		#	- if quotation then get chars until next non-escaped quotation. If there is any square bracket
		#		then read inside [] as function call
		
		#check if it's space/tab
		if { $char == " " || $char == "\t" } {
			return [getSpaceToken idx maxIdx line]
		}
		
		#braces
		if { $char == "\{" } {
			incr idx
			return [new Token $char [Token::TYPE_OPEN_BRACE]]
		} elseif { $char == "\}" } {
			incr idx
			return [new Token $char [Token::TYPE_CLOSE_BRACE]]
		}
		
		#another strings
		set quotation 0
		set tmpString ""
		set squareBracket 0
		set escape 0
		set brace 0
		
		while { $idx <= $maxIdx } {
			set char [string index $line $idx]
			
			if { !$escape } {			
				if { $char == "\"" } {
					set quotation [expr !$quotation]
				} elseif { $char ==	"\{" } {
				  #open brace inside current argument
				  incr brace
				} elseif { $char ==	"\}" } {
					if { $brace > 0 } {
						incr brace -1
					} elseif { !$quotation } {
						break
					}
				} elseif { $char ==	"\[" } {
					#square brackets inside braced strings should not be evaluated!
					if { !$brace } {
						#read function call
						append tmpString [getFunctionCallString idx maxIdx line]
						continue
					}
				} elseif { $char == " " || $char == "\t" } {
					if { !$quotation && !$brace } {
						#there is non-escaped space, so token is finished!
						break
					}			
				}
			}
			
			if { $char == "\\" } {
				set escape 1
			}	else {
				set escape 0
			}
		
			append tmpString $char
			incr idx
		}
		
		if { $quotation } {
			error "Non closed quotation in line '$line' !"		
		}
		
		if { $brace } {
			error "Non closed brace in line '$line' !"		
		}		
		
		return [new Token $tmpString [Token::TYPE_STRING]]	
	}
	
	proc getFunctionCallString { _idx _maxIdx _line } {
		upvar $_idx idx
		upvar $_maxIdx maxIdx
		upvar $_line line
					
		set squareBracketCounter 0		
		set functionCall ""
		set escape 0
		
		#get everything inside [ ]
		
		while { $idx <= $maxIdx } {
			set char [string index $line $idx]
			
			if { !$escape } {
				if { $char == "\[" } {
						incr squareBracketCounter
				} elseif { $char == "]" } {
						incr squareBracketCounter -1
				}
			}
			
			if { $char == "\\" } {
				set escape 1
			}	else {
				set escape 0
			}			
			
			append functionCall $char
			
			incr idx
			
			if { $squareBracketCounter == 0 } {
				break
			}			
		}
		
		if { $squareBracketCounter != 0 } {
			error "Non matched square bracket in line '$line'!"
		}
		
		set analyzedFunction [string range $functionCall 1 [expr [string length $functionCall] - 2] ]		
		set indentation 0 ;#ignore this
		
		analyzeLine analyzedFunction indentation
		
		return "\[$analyzedFunction\]"
	}
	
	#reads spaces as far as it can
	proc getSpaceToken { _idx _maxIdx _line } {
		upvar $_idx idx
		upvar $_maxIdx maxIdx
		upvar $_line line
		
		set tmp ""
		
		while { $idx <= $maxIdx } {
			set char [string index $line $idx]
			if { $char == " " || $char == "\t" } {
				append tmp $char			
				incr idx
			} else {
				break
			}		
		}
		return [new Token $tmp [Token::TYPE_EMPTY]]
	}
	
	proc isKnownCommand { _token } {
		return [expr [llength [info procs Analyzers::$_token]] == 1]
	}
}